//***************************************************************
// Filename: 	CLBWrite.c
//
// Brief:	This program writes a CLB structure 
//		in flash on the suzaku board
//
// Author:	Fredrik Gravdal, 2007, NTNU
//***************************************************************
#include <stdio.h>
#include <stdlib.h>
#include "CLBWrite.h"

typedef unsigned short __u16;

//changing little endian with big endian
#define ___swab16(x)                                                    \
        ({                                                              \
                __u16 __x = (x);                                        \
                ((__u16)(                                               \
                         (((__u16)(__x) & (__u16)0x00ffU) << 8) |       \
                         (((__u16)(__x) & (__u16)0xff00U) >> 8) ));     \
        })

#define flash_write(addr, b)                                    \
        (void)((*(volatile unsigned short *) (addr)) = ((b)))

#define flash_write_cmd(addr, b)                                        \
        (void)((*(volatile unsigned short *) (addr)) = (___swab16(b)))

//global variables to remember which sector that is in the buffer
int sectorInBuf=128;
unsigned int bufferAddr;
unsigned int oldSectorStartAddr;

//writes data to flash
void flash_program(unsigned long addr, unsigned short val)
{	
	flash_write_cmd(FLASH_START + 0xaaa, 0xaa);
	flash_write_cmd(FLASH_START + 0x554, 0x55);
	flash_write_cmd(FLASH_START + 0xaaa, 0xa0);
	flash_write(addr, val);
	while ((*(volatile unsigned short *) (addr)) != val);
}

//eraseing the sector in flash
void flash_sector_erase(int sectorAddr)
{
	printf("Erasing sector: %x\n", sectorAddr);

        flash_write_cmd(FLASH_START + 0xaaa, 0xaa);
        flash_write_cmd(FLASH_START + 0x554, 0x55);
        flash_write_cmd(FLASH_START + 0xaaa, 0x80);
        flash_write_cmd(FLASH_START + 0xaaa, 0xaa);
        flash_write_cmd(FLASH_START + 0x554, 0x55);

	flash_write_cmd(FLASH_START + sectorAddr, 0x30);

        while ((*(volatile unsigned short *) (FLASH_START + sectorAddr + FLASH_SECTOR_SIZE - 1)) != 0xFFFF);
}

//copying the sector in flash to buffer
void copy_sector_to_buf(unsigned short* buf, unsigned int secStartAddr)
{	
	printf("Copying sector: %x to buffer\n", secStartAddr);
        int i;
        for(i = 0; i < FLASH_SECTOR_SIZE; i+=2){
		buf[i/2] = (*(volatile unsigned short *) (secStartAddr + i));
	}
}

//copying the buffer to flash
void copy_buf_to_sector(unsigned short* buf, unsigned int secStartAddr)
{	
        int i;
	flash_sector_erase(secStartAddr-FLASH_START);
	printf("Copying buffer to sector: %x\n", secStartAddr);
	for(i = 0; i < FLASH_SECTOR_SIZE; i+=2){
		flash_program(secStartAddr + i, buf[i/2]);
	}
}

//updates the buffer, and writes it to the flash if needed
void update_flash(unsigned short* buf, unsigned int addr, unsigned int data, int finish){

	int sector = addr/FLASH_SECTOR_SIZE;
	unsigned int sectorStartAddr = sector*FLASH_SECTOR_SIZE;
	unsigned short tmpData1, tmpData2;

	if (finish == 1){
		copy_buf_to_sector(buf, oldSectorStartAddr);
	}
	else {
		//If the buffer is empty
		if(sectorInBuf == 128){
			printf("The buffer is empty\n");	
			copy_sector_to_buf(buf, sectorStartAddr);	
			oldSectorStartAddr = sectorStartAddr;
			sectorInBuf = sector;
			bufferAddr = addr;	
		}
		//If the address is in an another buffer
		else if(sectorInBuf != sector){
			printf("The address is in an another buffer\n");
			copy_buf_to_sector(buf, oldSectorStartAddr);
			copy_sector_to_buf(buf, sectorStartAddr);	
			oldSectorStartAddr = sectorStartAddr;
			sectorInBuf = sector;
			bufferAddr = addr;
		}
		
		tmpData2 = data;
		tmpData1 = data >>16;

		buf[(addr-sectorStartAddr)/2]  = tmpData1;
		buf[(addr-sectorStartAddr+2)/2]  = tmpData2;
	}
}

//FOR DEBUGGING 
//writes the flashbuffer to file
void dump_sector_buf_to_file(unsigned short* buf){

	printf("Dumping sector to sectorDump.bit \n");
	
	FILE * pFile;
  	pFile = fopen ("sectorDump.bit","wb");
  	
	int i;
	if (pFile!=NULL){	
		for(i = 0; i < FLASH_SECTOR_SIZE/2 ; i+=2){	
			fwrite(&buf[i/2], 2,1, pFile);
		}
	}
	else {
		printf("Error! Dumping sector to sectorDump.bit\n");
	}
	printf("Finished dumping sector to sectorDump.bit\n");
	fclose (pFile);
}

//main method. runs throug cols and rows in the input file and updates the flash
int main(int numArgs, char* argc[]){

	int startCol = atoi(argc[1]);
	int startRow = atoi(argc[2]);

	if(numArgs != 4){
		printf("Number of arguments does not mach. Should be 4, you entered: %d\n", numArgs);		
		return 0;	
	}
	else{		
		FILE * inFile;
		inFile = fopen (argc[3],"rb");
	 	if (inFile == NULL){printf("Error opening inFile\n"); }
		
		int endCol, endRow;
		unsigned int iData;

		fread(&iData, 4,1, inFile);
		endCol = iData;
		fread(&iData, 4,1, inFile);
		endRow = iData;
		
		endRow += startRow;
		endCol += startCol;

		printf("Writing from [%d] [%d] to [%d] [%d]\n",startRow,startCol,endRow,endCol);

		unsigned short buffer[FLASH_SECTOR_SIZE];
		clb oneCLB;
		clb CLBTable[endRow-startRow+1][endCol-startCol+1];

		fread(CLBTable, sizeof(oneCLB.iCLB[0]), sizeof(CLBTable)/sizeof(oneCLB.iCLB[0]), inFile);

		//writes the CLBTable to flash
		int k,col,row=0;
		for(col = startCol; col <= endCol; col++){
			for(row = startRow; row <= endRow; row++){

				printf("[%d] [%d]\n",col,row);
				oneCLB = CLBTable[row - startRow][col - startCol];

				 for (k=0;k<19*2;k+=2){	
					
					iData = oneCLB.iCLB[k];			
					update_flash(buffer, FLASH_START + FPGA_START + CLB_START + (row*4*2) + (col*101*4*19) + (k/2)*(101*4), iData, 0);
					
					iData = oneCLB.iCLB[k+1];			
					update_flash(buffer, FLASH_START + FPGA_START + CLB_START + (row*4*2) + (col*101*4*19) + (k/2)*(101*4) + 4, iData, 0);
				}
		
			}
		}

		//write buffer back to flash
		update_flash(buffer, 0, 0, 1);
		fclose (inFile);
	}
	return 0;
}
